<?php

namespace Drupal\adva\Plugin\adva;

use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;

use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Defines api layer for field access in AccessProviders plugins.
 */
class ReferenceAccessProvider extends AccessProvider implements ReferenceAccessProviderInterface {

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Create a new AccessProvider.
   *
   * @param array $configuration
   *   Plugin configuration.
   * @param string $plugin_id
   *   Unique plugin id.
   * @param array|mixed $plugin_definition
   *   Plugin instance definition.
   * @param Drupal\adva\Plugin\adva\AccessConsumerInterface $consumer
   *   Associated Access Consumer Instance.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The Entity Type manager to get entity storage.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, AccessConsumerInterface $consumer, EntityTypeManagerInterface $entity_type_manager) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $consumer);

    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition, AccessConsumerInterface $consumer) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $consumer,
      $container->get("entity_type.manager")
    );
  }

  /**
   * {@inheritdoc}
   */
  public static function getTargetType() {
    // Implementations should override this method.
    return NULL;
  }

  /**
   * {@inheritdoc}
   */
  public static function getReferenceFields($entity_type_id, $target_type_id) {
    $type_fields = &drupal_static(__METHOD__);
    if (!isset($type_fields[$entity_type_id]) || !isset($type_fields[$entity_type_id][$target_type_id])) {
      $type_fields[$entity_type_id][$target_type_id] = [];
      $entityBundleManager = \Drupal::service('entity_type.bundle.info');

      $bundles = $entityBundleManager->getBundleInfo($entity_type_id);
      foreach (array_keys($bundles) as $bundle) {
        $bundle_fields = self::getBundleReferenceFields($entity_type_id, $bundle, $target_type_id);
        $type_fields[$entity_type_id][$target_type_id] = array_merge($type_fields[$entity_type_id][$target_type_id], $bundle_fields);
      }
    }
    return $type_fields[$entity_type_id][$target_type_id];
  }

  /**
   * {@inheritdoc}
   */
  public static function getBundleReferenceFields($entity_type_id, $bundle, $target_type_id) {
    $bundle_fields = &drupal_static(__METHOD__);
    if (!isset($bundle_fields[$entity_type_id])
      || !isset($bundle_fields[$entity_type_id][$bundle])
      || !isset($bundle_fields[$entity_type_id][$bundle][$target_type_id])) {
      // Create an empty array, to ensure there is a value to return.
      $bundle_fields[$entity_type_id][$bundle][$target_type_id] = [];

      $reference_field_types = [
        'entity_reference',
        'entity_reference_revisions',
        'field_collection',
      ];

      $entityFieldManager = \Drupal::service('entity_field.manager');
      $fields = $entityFieldManager->getFieldDefinitions($entity_type_id, $bundle);
      foreach ($fields as $field_name => $field) {
        $field_type = method_exists($field, 'getType') ? $field->getType() : NULL;
        $target_type = method_exists($field, 'getSetting') ? $field->getSetting('target_type') : NULL;
        if (in_array($field_type, $reference_field_types) && $target_type === $target_type_id) {
          $bundle_fields[$entity_type_id][$bundle][$target_type_id][$field_name] = $field;
        }
      }
    }
    return $bundle_fields[$entity_type_id][$bundle][$target_type_id];
  }

  /**
   * {@inheritdoc}
   */
  public function getEntityTypeManager() {
    return $this->entityTypeManager;
  }

  /**
   * {@inheritdoc}
   */
  public function getEnabledBundleFields($entity_type_id, $bundle = NULL) {
    if ($bundle === NULL) {
      $bundle = $entity_type_id;
    }

    $enabled_field_names = $this->configuration['enabled_fields'];
    $bundle_fields = static::getBundleReferenceFields($entity_type_id, $bundle, static::getTargetType());
    $bundle_field_names = array_keys($bundle_fields);

    return array_intersect($bundle_field_names, $enabled_field_names);
  }

  /**
   * {@inheritdoc}
   */
  public function getReferencedItems(EntityInterface $entity) {
    static $item_ids = [];

    if (!$entity) {
      return [];
    }

    $id = $entity->id();
    $type = $entity->getEntityType();
    $type_id = $type->id();
    if (!isset($item_ids[$type_id]) || !isset($item_ids[$type_id][$id])) {
      // Ensure there will be a returnable value.
      $item_ids[$type_id][$id] = [];

      // Get list of fields.
      $bundle = $entity->bundle();
      $fields = $this->getEnabledBundleFields($type_id, $bundle);

      // Get values for each field.
      foreach ($fields as $field_name) {

        if (($field = $entity->get($field_name)) && ($items = $field->getValue())) {
          // Filter out empty values from items list.
          $items = array_filter($items);
          $type_key = ($field->getFieldDefinition()->getType() === 'field_collection') ? 'value' : 'target_id';
          $values = array_column($items, $type_key);
          $item_ids[$type_id][$id] = array_merge($item_ids[$type_id][$id], $values);
        }
      }

      // Filter out duplicate values, from itemss present in multiple fields.
      $item_ids[$type_id][$id] = array_unique($item_ids[$type_id][$id]);
      // Null values are inserted with loading forms. Filter them out.
      $item_ids[$type_id][$id] = array_filter($item_ids[$type_id][$id]);
    }
    return isset($item_ids[$type_id][$id]) ? $item_ids[$type_id][$id] : [];
  }

  /**
   * {@inheritdoc}
   */
  public static function appliesToType(EntityTypeInterface $entityType) {
    if (!$entityType->isSubclassOf('\Drupal\Core\Entity\FieldableEntityInterface')) {
      return FALSE;
    }

    return (count(static::getReferenceFields($entityType->id(), static::getTargetType())) > 0);
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigForm(array $form, FormStateInterface $form_state) {
    $consumer = $this->getConsumer();
    $entity_type_id = $consumer->getEntityTypeId();
    $form['enabled_fields'] = [
      '#type' => 'checkboxes',
      '#title' => $this->t('Enabled Fields'),
      '#description' => $this->t('Choose the fields to use when checking access.'),
      '#default_value' => $this->configuration['enabled_fields'] ?: [],
      '#parents' => [
        'consumers',
        $consumer->getPluginId(),
        'config',
        'providers',
        $this->getPluginId(),
        'enabled_fields',
      ],
      '#options' => [],
    ];

    $available_fields = static::getReferenceFields($entity_type_id, static::getTargetType());
    foreach ($available_fields as $field_name => $field) {
      $form['enabled_fields']['#options'][$field_name] = $field->label();
    }
    if (count($form['enabled_fields']['#options']) === 0) {
      $form['enabled_fields']['#access'] = FALSE;
      $entityTypeManager = $this->getEntityTypeManager();
      $context = [
        '%target_type' => $entityTypeManager->getDefinition(static::getTargetType())->getLabel(),
        '%label' => $this->getLabel(),
        '%entity_type' => $entityTypeManager->getDefinition($entity_type_id)->getLabel(),
      ];
      $form['message'] = [
        '#type' => 'item',
        '#title' => $this->t('Enabled Fields: %target_type', $context),
        '#markup' => $this->t('There are no available %target_type reference fields. To use %label, please add a %target_type field to the %entity_type entity.', $context),
      ];
    }
    elseif (count($form['enabled_fields']['#default_value']) > 0) {
      $form['#open'] = TRUE;
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigForm(array &$form, FormStateInterface $form_state) {
    $consumer = $this->getConsumer();
    $values = $form_state->getValues();
    $this->configuration['enabled_fields'] = array_filter($values['consumers'][$consumer->getPluginId()]['config']['providers'][$this->getPluginId()]['enabled_fields']);
  }

}
